from ..builders import OutputBuilder
from ..builders.utils import ensure_parent_modules
from ..outputs.toml import TOMLOutput
from ..utils.verbose import log


class InvenioRecordMetadataAlembicPoetryBuilder(OutputBuilder):
    TYPE = 'invenio_record_metadata_alembic_poetry'

    def finish(self):
        super().finish()

        output: TOMLOutput = self.builder.get_output(
            'toml',
            'pyproject.toml'
        )
        output.set("tool.poetry.plugins.'invenio_db.alembic'",
                   self.settings.python.record_schema_metadata_alembic,
                   f'{self.settings.package}:alembic'
                   )

        python_path = self.settings.package_path / 'alembic' / '__init__.py'
        # create parent modules if they do not exist
        ensure_parent_modules(
            self.builder,
            python_path,
            max_depth=len(python_path.parts)
        )

        # and create empty __init__.py
        init_builder = self.builder.get_output(
            'python',
            python_path
        )
        if init_builder.created:
            log(log.INFO, f"""Do not forget to run:
    
    # if the initial database does not exist yet 
    invenio db init
    
    # if the tables do not exist yet (that is, after invenio db init); you have to manually remove 
    # {self.settings.python.record_metadata_table_name} and its versioned counterpart, otherwise
    # alembic below will not work !
    invenio db create 
            
    # create the branch
    invenio alembic revision "Create {self.settings.python.record_schema_metadata_alembic} branch."  -b {self.settings.python.record_schema_metadata_alembic} -p dbdbc1b19cf2 --empty
    
    # apply the branch
    invenio alembic upgrade heads
    
    # initial revision
    invenio alembic revision "Initial revision." -b {self.settings.python.record_schema_metadata_alembic}
    
    # inspect the generated file and add import sqlalchemy_utils (invenio template does not contain it
    # remove length=16 from UUIDType(length=16), replace Text() with sa.Text()
    
    # create db tables
    invenio alembic upgrade heads
            """)



from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.invenio.invenio_record_schema import InvenioRecordSchemaBuilder
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.stack import ModelBuilderStack
from oarepo_model_builder.utils.deepmerge import deepmerge


class DatePreprocessor(PropertyPreprocessor):
    TYPE = 'date'

    @process(model_builder=JSONSchemaBuilder,
             path='**/properties/*',
             condition=lambda current, stack: current.type == 'date')
    def modify_date_jsonschema(self, data, stack: ModelBuilderStack, **kwargs):
        data['type'] = 'string'
        data.setdefault('format', 'date')
        return data

    @process(model_builder=InvenioRecordSchemaBuilder,
             path='**/properties/*',
             condition=lambda current, stack: current.type == 'date')
    def modify_date_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        deepmerge(
            data.setdefault('oarepo:marshmallow', {}),
            {
                'class': 'ma_fields.Date'
                # TODO: validators, such as minDate, maxDate, minDateExclusive, maxDateExclusive
            }
        )
        return data



import re
from collections import OrderedDict, namedtuple
from functools import cached_property
from typing import List, Iterable

"""
A path is a sequence of names separated by '/' with an optional 
condition applied to the subtree.

The condition is a function taking current:PathCondition as a single parameter
which can be used for example as: 
current.a.b == "data" or current["a"]["b"] == "data" 

Any expression with the current should return an iterator or list of results (matched subtrees).
If any is truish, the condition is interpreted as matching.
"""

JSONPathRecord = namedtuple('JSONPathRecord', 'path, condition, value')


class JSONPaths:
    def __init__(self):
        self.path_regex_list = []
        self.path_to_index = {}
        self.paths: List[List[JSONPathRecord]] = []

    def register(self, path, condition=None, value=None):
        """
        paths should be registered from longest to shortest

        :param path:        a path to register
        :param condition:   a condition applied to the node on the path
        :param value:       the value to return if path & condition match
        """

        if path not in self.path_to_index:
            self.path_to_index[path] = len(self.path_to_index)
            self.paths.append([])
            self.path_regex_list.append(path_to_regex(path))
        path_locators = self.paths[self.path_to_index[path]]
        path_locators.append(JSONPathRecord(
            path=path,
            condition=condition,
            value=value
        ))

    @cached_property
    def path_regex(self):
        return re.compile('|'.join(f'({x})' for x in self.path_regex_list))

    def match(self, path=None, subtree=None, extra_data=None):
        """
        Matches a path and subtree against stored paths. Returns iterator of matched values

        :param path:        the path that should match
        :param subtree:     teh subtree against which to match stored conditions
        :return:            iterator of matched values
        """
        match = self.path_regex.match(path)
        if not match:
            return None
        for idx, grp in enumerate(match.groups()):
            if grp:
                matched = False
                for rec in self.paths[idx]:
                    if rec.condition:
                        condition_result = rec.condition(PathCondition(subtree), **(extra_data or {}))
                        if isinstance(condition_result, Iterable):
                            condition_result = any(condition_result)
                        if condition_result:
                            matched = True
                            yield rec.value
                    else:
                        matched = True
                        yield rec.value

                if matched:
                    break


def path_to_regex(path):
    split_path = [x for x in re.split('(/)', path) if x]

    def fragment_to_regex(f):
        if f == '**':
            return '.+'
        return f.replace('*', '[^/]+')

    return ''.join(fragment_to_regex(x) for x in split_path)


class PathCondition:
    def __init__(self, start=None, subtree_list=()):
        self._subtree_list = [*subtree_list]
        if start:
            self._subtree_list.append(start)

    def _apply(self, p, subtree_list):
        if subtree_list:
            if p == '*':
                for subtree in subtree_list:
                    itr = []
                    if isinstance(subtree, dict):
                        itr = subtree.values()
                    elif isinstance(subtree, list):
                        itr = subtree
                    yield from itr
            elif p == '**':
                first_level = list(self._apply('*', subtree_list))
                yield from first_level
                yield from self._apply('**', first_level)
            else:
                for subtree in subtree_list:
                    # match the path element and descend if found
                    if isinstance(subtree, dict):
                        if p in subtree:
                            yield subtree[p]
                    elif isinstance(subtree, list):
                        if isinstance(p, int) and 0 <= p < len(subtree):
                            yield subtree[p]

    def __getattr__(self, item):
        try:
            item = int(item)
        except:
            pass
        return PathCondition(subtree_list=list(self._apply(item, self._subtree_list)))

    def __getitem__(self, item):
        return PathCondition(subtree_list=list(self._apply(item, self._subtree_list)))

    def __eq__(self, other):
        if isinstance(other, PathCondition):
            # TODO: deep equals maybe
            return [subtree for subtree in self._subtree_list for o in other._subtree_list if subtree == o]

        # TODO: deep equals maybe
        return [subtree for subtree in self._subtree_list if subtree == other]



from oarepo_model_builder.builders import process
from oarepo_model_builder.builders.python import PythonBuilder
from oarepo_model_builder.builders.utils import ensure_parent_modules
from oarepo_model_builder.stack import ModelBuilderStack


class PythonStructureBuilder(PythonBuilder):
    TYPE = 'python_structure'

    @process('/model')
    def model(self, stack: ModelBuilderStack):
        yield
        package_path = self.settings.package_path

        ensure_parent_modules(
            self.builder,
            self.builder.output_dir.joinpath(package_path / '__init__.py'),
            max_depth=len(package_path.parts)
        )



from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.builders.mapping import MappingBuilder
from oarepo_model_builder.invenio.invenio_record_schema import InvenioRecordSchemaBuilder
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.utils.deepmerge import deepmerge
from oarepo_model_builder.stack import ModelBuilderStack


class TextKeywordPreprocessor(PropertyPreprocessor):
    TYPE = 'text_keyword'
    #
    # type='fulltext' in model
    #

    @process(model_builder=JSONSchemaBuilder,
             path='**/properties/*',
             condition=lambda current, stack: current.type == 'fulltext')
    def modify_fulltext_schema(self, data, stack: ModelBuilderStack, **kwargs):
        data['type'] = 'string'
        return data

    @process(model_builder=MappingBuilder,
             path='**/properties/*',
             condition=lambda current, stack: current.type == 'fulltext')
    def modify_fulltext_mapping(self, data, stack: ModelBuilderStack, **kwargs):
        data['type'] = 'text'
        return data

    @process(model_builder=InvenioRecordSchemaBuilder,
             path='**/properties/*',
             condition=lambda current, stack: current.type == 'fulltext')
    def modify_fulltext_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        data['type'] = 'string'
        return data

    #
    # type='keyword' in model
    #

    @process(model_builder=JSONSchemaBuilder,
             path='**/properties/*',
             condition=lambda current, stack: current.type == 'keyword')
    def modify_keyword_schema(self, data, stack: ModelBuilderStack, **kwargs):
        data['type'] = 'string'
        return data

    @process(model_builder=MappingBuilder,
             path='**/properties/*',
             condition=lambda current, stack: current.type == 'keyword')
    def modify_keyword_mapping(self, data, stack: ModelBuilderStack, **kwargs):
        data['type'] = 'keyword'
        deepmerge(
            data.setdefault('oarepo:mapping', {}),
            {
                'ignore_above': self.settings['elasticsearch']['keyword-ignore-above']
            })
        return data

    @process(model_builder=InvenioRecordSchemaBuilder,
             path='**/properties/*',
             condition=lambda current, stack: current.type == 'keyword')
    def modify_keyword_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        data['type'] = 'string'
        return data

    #
    # type='fulltext+keyword' in model
    #

    @process(model_builder=JSONSchemaBuilder,
             path='**/properties/*',
             condition=lambda current, stack: current.type == 'fulltext+keyword')
    def modify_fulltext_keyword_schema(self, data, stack: ModelBuilderStack, **kwargs):
        data['type'] = 'string'
        return data

    @process(model_builder=MappingBuilder,
             path='**/properties/*',
             condition=lambda current, stack: current.type == 'fulltext+keyword')
    def modify_fulltext_keyword_mapping(self, data, stack: ModelBuilderStack, **kwargs):
        data['type'] = 'text'
        deepmerge(
            data.setdefault('oarepo:mapping', {}),
            {
                'fields': {
                    'keyword': {
                        'type': 'keyword',
                        'ignore_above': self.settings['elasticsearch']['keyword-ignore-above']
                    }
                }
            }, [])
        return data

    @process(model_builder=InvenioRecordSchemaBuilder,
             path='**/properties/*',
             condition=lambda current, stack: current.type == 'fulltext+keyword')
    def modify_fulltext_keyword_marshmallow(self, data, stack: ModelBuilderStack, **kwargs):
        data['type'] = 'string'
        return data
